# -*- coding: utf-8 -*-
# MIT License: Copyright (c) 2022, Marie Roald.
# See the LICENSE file in the root directory for full license text.

__author__ = """Marie Roald"""
__email__ = "roald.marie@gmail.com"
__version__ = "0.2.0"


from . import coupled_matrices, data, decomposition, penalties, random
